#include "include/parsertl/debug.hpp"
#include "include/parsertl/generator.hpp"
#include <iostream>
#include "../lexertl/include/lexertl/memory_file.hpp"

#include "include/parsertl/lookup.hpp"
#include "include/parsertl/search.hpp"

void build_c()
{
	parsertl::rules grules_;
	parsertl::state_machine gsm_;

	grules_.token("IDENTIFIER CONSTANT STRING_LITERAL SIZEOF");
	grules_.token("PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP "
		"EQ_OP NE_OP");
	grules_.token("AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN "
		"ADD_ASSIGN");
	grules_.token("SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN");
	grules_.token("XOR_ASSIGN OR_ASSIGN TYPE_NAME");
	grules_.token("TYPEDEF EXTERN STATIC AUTO REGISTER");
	grules_.token("CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE CONST "
		"VOLATILE VOID");
	grules_.token("STRUCT UNION ENUM ELIPSIS RANGE");
	grules_.token("CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE "
		"BREAK RETURN");

	grules_.push("accept", "file");

	grules_.push("primary_expr", "identifier"
		"| CONSTANT"
		"| STRING_LITERAL"
		"| '(' expr ')'");
	grules_.push("postfix_expr", "primary_expr"
		"| postfix_expr '[' expr ']'"
		"| postfix_expr '(' ')'"
		"| postfix_expr '(' argument_expr_list ')'"
		"| postfix_expr '.' identifier"
		"| postfix_expr PTR_OP identifier"
		"| postfix_expr INC_OP"
		"| postfix_expr DEC_OP");
	grules_.push("argument_expr_list", "assignment_expr"
		"| argument_expr_list ',' assignment_expr");
	grules_.push("unary_expr", "postfix_expr"
		"| INC_OP unary_expr"
		"| DEC_OP unary_expr"
		"| unary_operator cast_expr"
		"| SIZEOF unary_expr"
		"| SIZEOF '(' type_name ')'");
	grules_.push("unary_operator", "'&' "
		"| '*'"
		"| '+'"
		"| '-'"
		"| '~'"
		"| '!'");
	grules_.push("cast_expr", "unary_expr"
		"| '(' type_name ')' cast_expr");
	grules_.push("multiplicative_expr", "cast_expr"
		"| multiplicative_expr '*' cast_expr"
		"| multiplicative_expr '/' cast_expr"
		"| multiplicative_expr '%' cast_expr");
	grules_.push("additive_expr", "multiplicative_expr"
		"| additive_expr '+' multiplicative_expr"
		"| additive_expr '-' multiplicative_expr");
	grules_.push("shift_expr", "additive_expr"
		"| shift_expr LEFT_OP additive_expr"
		"| shift_expr RIGHT_OP additive_expr");
	grules_.push("relational_expr", "shift_expr"
		"| relational_expr '<' shift_expr"
		"| relational_expr '>' shift_expr"
		"| relational_expr LE_OP shift_expr"
		"| relational_expr GE_OP shift_expr");
	grules_.push("equality_expr", "relational_expr"
		"| equality_expr EQ_OP relational_expr"
		"| equality_expr NE_OP relational_expr");
	grules_.push("and_expr", "equality_expr"
		"| and_expr '&' equality_expr");
	grules_.push("exclusive_or_expr", "and_expr"
		"| exclusive_or_expr '^' and_expr");
	grules_.push("inclusive_or_expr", "exclusive_or_expr"
		"| inclusive_or_expr '|' exclusive_or_expr");
	grules_.push("logical_and_expr", "inclusive_or_expr"
		"| logical_and_expr AND_OP inclusive_or_expr");
	grules_.push("logical_or_expr", "logical_and_expr"
		"| logical_or_expr OR_OP logical_and_expr");
	grules_.push("conditional_expr", "logical_or_expr"
		"| logical_or_expr '?' logical_or_expr ':' conditional_expr");
	grules_.push("assignment_expr", "conditional_expr"
		"| unary_expr assignment_operator assignment_expr");
	grules_.push("assignment_operator", "'=' "
		"| MUL_ASSIGN"
		"| DIV_ASSIGN"
		"| MOD_ASSIGN"
		"| ADD_ASSIGN"
		"| SUB_ASSIGN"
		"| LEFT_ASSIGN"
		"| RIGHT_ASSIGN"
		"| AND_ASSIGN"
		"| XOR_ASSIGN"
		"| OR_ASSIGN");
	grules_.push("expr", "assignment_expr"
		"| expr ',' assignment_expr");
	grules_.push("constant_expr", "conditional_expr");
	grules_.push("declaration", "declaration_specifiers ';' "
		"| declaration_specifiers init_declarator_list ';'");
	grules_.push("declaration_specifiers", "storage_class_specifier"
		"| storage_class_specifier declaration_specifiers"
		"| type_specifier"
		"| type_specifier declaration_specifiers");
	grules_.push("init_declarator_list", "init_declarator"
		"| init_declarator_list ',' init_declarator");
	grules_.push("init_declarator", "declarator"
		"| declarator '=' initializer");
	grules_.push("storage_class_specifier", "TYPEDEF"
		"| EXTERN"
		"| STATIC"
		"| AUTO"
		"| REGISTER");
	grules_.push("type_specifier", "CHAR"
		"| SHORT"
		"| INT"
		"| LONG"
		"| SIGNED"
		"| UNSIGNED"
		"| FLOAT"
		"| DOUBLE"
		"| CONST"
		"| VOLATILE"
		"| VOID"
		"| struct_or_union_specifier"
		"| enum_specifier"
		"| TYPE_NAME");
	grules_.push("struct_or_union_specifier",
		"struct_or_union identifier '{' struct_declaration_list '}' "
		"| struct_or_union '{' struct_declaration_list '}'"
		"| struct_or_union identifier");
	grules_.push("struct_or_union", "STRUCT"
		"| UNION");
	grules_.push("struct_declaration_list", "struct_declaration"
		"| struct_declaration_list struct_declaration");
	grules_.push("struct_declaration",
		"type_specifier_list struct_declarator_list ';'");
	grules_.push("struct_declarator_list", "struct_declarator"
		"| struct_declarator_list ',' struct_declarator");
	grules_.push("struct_declarator", "declarator"
		"| ':' constant_expr"
		"| declarator ':' constant_expr");
	grules_.push("enum_specifier", "ENUM '{' enumerator_list '}' "
		"| ENUM identifier '{' enumerator_list '}'"
		"| ENUM identifier");
	grules_.push("enumerator_list", "enumerator"
		"| enumerator_list ',' enumerator");
	grules_.push("enumerator", "identifier"
		"| identifier '=' constant_expr");
	grules_.push("declarator", "declarator2 "
		"| pointer declarator2");
	grules_.push("declarator2", "identifier "
		"| '(' declarator ')'"
		"| declarator2 '[' ']'"
		"| declarator2 '[' constant_expr ']'"
		"| declarator2 '(' ')'"
		"| declarator2 '(' parameter_type_list ')'"
		"| declarator2 '(' parameter_identifier_list ')'");
	grules_.push("pointer", "'*'"
		"| '*' type_specifier_list"
		"| '*' pointer"
		"| '*' type_specifier_list pointer");
	grules_.push("type_specifier_list", "type_specifier"
		"| type_specifier_list type_specifier");
	grules_.push("parameter_identifier_list", "identifier_list"
		"| identifier_list ',' ELIPSIS");
	grules_.push("identifier_list", "identifier"
		"| identifier_list ',' identifier");
	grules_.push("parameter_type_list", "parameter_list"
		"| parameter_list ',' ELIPSIS");
	grules_.push("parameter_list", "parameter_declaration"
		"| parameter_list ',' parameter_declaration");
	grules_.push("parameter_declaration", "type_specifier_list declarator "
		"| type_name");
	grules_.push("type_name", "type_specifier_list"
		"| type_specifier_list abstract_declarator");
	grules_.push("abstract_declarator", "pointer"
		"| abstract_declarator2"
		"| pointer abstract_declarator2");
	grules_.push("abstract_declarator2", "'(' abstract_declarator ')' "
		"| '[' ']'"
		"| '[' constant_expr ']'"
		"| abstract_declarator2 '[' ']'"
		"| abstract_declarator2 '[' constant_expr ']'"
		"| '(' ')'"
		"| '(' parameter_type_list ')'"
		"| abstract_declarator2 '(' ')'"
		"| abstract_declarator2 '(' parameter_type_list ')'");
	grules_.push("initializer", "assignment_expr"
		"| '{' initializer_list '}'"
		"| '{' initializer_list ',' '}'");
	grules_.push("initializer_list", "initializer"
		"| initializer_list ',' initializer");
	grules_.push("statement", "labeled_statement"
		"| compound_statement"
		"| expression_statement"
		"| selection_statement"
		"| iteration_statement"
		"| jump_statement");
	grules_.push("labeled_statement", "identifier ':' statement "
		"| CASE constant_expr ':' statement"
		"| DEFAULT ':' statement");
	grules_.push("compound_statement", "'{' '}' "
		"| '{' statement_list '}'"
		"| '{' declaration_list '}'"
		"| '{' declaration_list statement_list '}'");
	grules_.push("declaration_list", "declaration"
		"| declaration_list declaration");
	grules_.push("statement_list", "statement"
		"| statement_list statement");
	grules_.push("expression_statement", "';' "
		"| expr ';'");
	grules_.push("selection_statement", "IF '(' expr ')' statement "
		"| IF '(' expr ')' statement ELSE statement"
		"| SWITCH '(' expr ')' statement");
	grules_.push("iteration_statement", "WHILE '(' expr ')' statement "
		"| DO statement WHILE '(' expr ')' ';' "
		"| FOR '(' ';' ';' ')' statement"
		"| FOR '(' ';' ';' expr ')' statement"
		"| FOR '(' ';' expr ';' ')' statement"
		"| FOR '(' ';' expr ';' expr ')' statement"
		"| FOR '(' expr ';' ';' ')' statement"
		"| FOR '(' expr ';' ';' expr ')' statement"
		"| FOR '(' expr ';' expr ';' ')' statement"
		"| FOR '(' expr ';' expr ';' expr ')' statement");
	grules_.push("jump_statement", "GOTO identifier ';' "
		"| CONTINUE ';'"
		"| BREAK ';'"
		"| RETURN ';'"
		"| RETURN expr ';'");
	grules_.push("file", "external_definition "
		"| file external_definition");
	grules_.push("external_definition", "function_definition "
		"| declaration");
	grules_.push("function_definition", "declarator function_body "
		"| declaration_specifiers declarator function_body");
	grules_.push("function_body", "compound_statement "
		"| declaration_list compound_statement");
	grules_.push("identifier", "IDENTIFIER");

	std::string warnings_;

	parsertl::generator::build(grules_, gsm_, &warnings_);
	std::cout << "Warnings: " << warnings_ << '\n';
}

void run_calc()
{
	parsertl::rules grules_;
	parsertl::state_machine gsm_;

	grules_.token("INTEGER");
	grules_.left("'+' '-'");
	grules_.left("'*' '/'");
	grules_.precedence("UMINUS");

	// Calculator from flex & bison book.
	grules_.start("start");

	const std::size_t add_idx_ = grules_.push("exp", "exp '+' exp");
	const std::size_t sub_idx_ = grules_.push("exp", "exp '-' exp");
	const std::size_t mul_idx_ = grules_.push("exp", "exp '*' exp");
	const std::size_t div_idx_ = grules_.push("exp", "exp '/' exp");
	const std::size_t umin_idx_ = grules_.push("exp", "'-' exp %prec UMINUS");
	const std::size_t int_idx_ = grules_.push("exp", "INTEGER");

	grules_.push("exp", "'(' exp ')'");
	grules_.push("start", "exp");

	std::string warnings_;

	parsertl::generator::build(grules_, gsm_, &warnings_);
	std::cout << "Warnings: " << warnings_ << '\n';

	lexertl::rules lrules_;
	lexertl::state_machine lsm_;

	lrules_.push("[+]", grules_.token_id("'+'"));
	lrules_.push("-", grules_.token_id("'-'"));
	lrules_.push("[*]", grules_.token_id("'*'"));
	lrules_.push("[/]", grules_.token_id("'/'"));
	lrules_.push("\\d+", grules_.token_id("INTEGER"));
	lrules_.push("[(]", grules_.token_id("'('"));
	lrules_.push("[)]", grules_.token_id("')'"));
	lrules_.push("\\s+", lsm_.skip());
	lexertl::generator::build(lrules_, lsm_);

	std::string expr_("1 + 2 * -3");
	lexertl::citerator iter_(expr_.c_str(), expr_.c_str() + expr_.size(), lsm_);
	parsertl::match_results gresults_(iter_->id, gsm_);
	typedef parsertl::token<lexertl::citerator> token;
	token::token_vector productions_;
	std::stack<int> stack_;

	while (gresults_.entry.action != parsertl::accept)
	{
		switch (gresults_.entry.action)
		{
		case parsertl::error:
			throw std::runtime_error("Parser error");
			break;
		case parsertl::reduce:
		{
			const std::size_t id_ = gresults_.reduce_id();

			if (id_ == add_idx_)
			{
				const int rhs_ = stack_.top();

				stack_.pop();
				stack_.top() = stack_.top() + rhs_;
			}
			else if (id_ == sub_idx_)
			{
				const int rhs_ = stack_.top();

				stack_.pop();
				stack_.top() = stack_.top() - rhs_;
			}
			else if (id_ == mul_idx_)
			{
				const int rhs_ = stack_.top();

				stack_.pop();
				stack_.top() = stack_.top() * rhs_;
			}
			else if (id_ == div_idx_)
			{
				const int rhs_ = stack_.top();

				stack_.pop();
				stack_.top() = stack_.top() / rhs_;
			}
			else if (id_ == umin_idx_)
			{
				stack_.top() *= -1;
			}
			else if (id_ == int_idx_)
			{
				stack_.push(atoi(gresults_.dollar(0, gsm_, productions_).
					first));
			}

			break;
		}
		}

		parsertl::lookup(iter_, gsm_, gresults_, productions_);
	}

	std::cout << expr_ << " = " << stack_.top() << '\n';
	parsertl::debug::dump(grules_, std::cout);
}

void find_uninit()
{
	typedef parsertl::basic_rules<char, unsigned char> grules;
	typedef parsertl::basic_state_machine<unsigned char> gstate_machine;
	typedef parsertl::basic_generator<grules, gstate_machine> ggenerator;
	typedef lexertl::basic_state_machine<char, unsigned char> lstate_machine;
	typedef lexertl::basic_rules<char, char, unsigned char> lrules;
	typedef lexertl::basic_generator<lrules, lstate_machine> lgenerator;
	typedef lexertl::iterator<const char *, lstate_machine,
		lexertl::match_results<const char *, unsigned char> > citerator;
	grules grules_;
	gstate_machine gsm_;
	lrules lrules_;
	lstate_machine lsm_;

	try
	{
		grules_.token("Bool Char Name NULLPTR Number String Type");
		grules_.push("start", "decl");
		grules_.push("decl", "Type list ';'");
		grules_.push("list", "item "
			"| list ',' item");
		const std::size_t uninit_idx_ = grules_.push("item", "Name");
		const std::size_t init_idx_ = grules_.push("item", "Name '=' value");
		grules_.push("value", "Bool "
			"| Char "
			"| Number "
			"| NULLPTR "
			"| String");
		ggenerator::build(grules_, gsm_);
		lrules_.insert_macro("NAME", "[_A-Za-z][_0-9A-Za-z]*");
		lrules_.push("=", grules_.token_id("'='"));
		lrules_.push(",", grules_.token_id("','"));
		lrules_.push(";", grules_.token_id("';'"));
		lrules_.push("true|TRUE|false|FALSE", grules_.token_id("Bool"));
		lrules_.push("nullptr", grules_.token_id("NULLPTR"));
		lrules_.push("BOOL|BSTR|BYTE|COLORREF|D?WORD|DWORD_PTR|DROPEFFECT|HACCEL|HANDLE|"
			"HBITMAP|HBRUSH|HCRYPTHASH|HCRYPTKEY|HCRYPTPROV|HCURSOR|HDBC|HICON|"
			"HINSTANCE|HMENU|HMODULE|HSTMT|HTREEITEM|HWND|LPARAM|LPCTSTR|"
			"LPDEVMODE|POSITION|SDWORD|SQLHANDLE|SQLINTEGER|SQLSMALLINT|UINT|"
			"U?INT_PTR|UWORD|WPARAM|"
			"bool|(unsigned\\s+)?char|double|float|"
			"(unsigned\\s+)?int((32|64)_t)?|long|size_t|"
			"{NAME}(\\s*::\\s*{NAME})*(\\s*[*])+", grules_.token_id("Type"));
		lrules_.push("{NAME}", grules_.token_id("Name"));
		lrules_.push("-?\\d+([.]\\d+)?", grules_.token_id("Number"));
		lrules_.push("'([^']|\\\\')*'", grules_.token_id("Char"));
		lrules_.push("[\"]([^\"]|\\\\[\"])*[\"]", grules_.token_id("String"));
		lrules_.push("[ \t\r\n]+|[/][/].*|[/][*](.|\n)*?[*][/]", lrules_.skip());
		lgenerator::build(lrules_, lsm_);

		lexertl::memory_file mf_("D:\\Ben\\Dev\\scratchpad\\AcAutoPostEquitracDlg.h");
		citerator iter_(mf_.data(), mf_.data() + mf_.size(), lsm_);
		citerator end_;
		std::set<unsigned char> set_;

		do
		{
			if (parsertl::search(iter_, end_, gsm_, &set_))
				if (set_.find((unsigned char)uninit_idx_) != set_.end())
					std::cout << std::string(iter_->first, end_->first) << '\n';

			iter_ = end_;
		} while (iter_->id != 0);
	}
	catch (const std::exception &e)
	{
		std::cout << e.what() << '\n';
	}
}

int main()
{
	try
	{
//		find_uninit();
		run_calc();
//		build_c();
	}
	catch (const std::exception &e)
	{
		std::cout << e.what() << '\n';
	}

	return 0;
}
